import os
import logging
import ssl
import paho.mqtt.client as mqtt

try:
    TLS_VERSION = ssl.PROTOCOL_TLS
except AttributeError:
    TLS_VERSION = ssl.PROTOCOL_TLSv1

logger = logging.getLogger(__name__)


class SimpleMQTTHost:

    CONFIGURABLE_OPTIONS = [
        "mqtt_client_id",
        "mqtt_username",
        "mqtt_password",
        "mqtt_host",
        "mqtt_port",
        "mqtt_tls_cacert",
        "mqtt_tls_certfile",
        "mqtt_tls_keyfile",
        "discovery_prefix",
        "node_id"
    ]

    def __init__(self):
        self.mqtt_client = None
        self.discovery_prefix = "homeassistant"
        self.node_id = None

        self.mqtt_client_id = ""

        self.mqtt_username = None
        self.mqtt_password = None

        self.mqtt_host = "localhost"
        self.mqtt_port = 1883

        self.mqtt_tls_cacert = None
        self.mqtt_tls_certfile = None
        self.mqtt_tls_keyfile = None

    def add_device(self, device):
        device.connect(self.mqtt_client, self.discovery_prefix, self.node_id)

    def start(self, block=True):
        if self.mqtt_client is None:
            self.mqtt_client = mqtt.Client(
                client_id=self.mqtt_client_id,
                clean_session=self.mqtt_client_id == ""
            )

            if self.mqtt_username is not None:
                self.mqtt_client.username_pw_set(
                    self.mqtt_username, self.mqtt_password
                )

            if self.mqtt_tls_cacert is None and self.mqtt_tls_certfile is not None:
                logger.warning("mqtt_tls_cacert not set, ignoring mqtt_tls_certfile setting")
            elif self.mqtt_tls_cacert is not None:
                self.mqtt_client.tls_set(
                    ca_certs=self.mqtt_tls_cacert,
                    keyfile=self.mqtt_tls_keyfile,
                    certfile=self.mqtt_tls_certfile,
                    tls_version=TLS_VERSION
                )

        self.mqtt_client.connect(self.mqtt_host, self.mqtt_port)

        if block:
            self.mqtt_client.loop_forever()
        else:
            self.mqtt_client.loop_start()

    def stop(self):
        if self.mqtt_client is not None:
            self.mqtt_client.loop_stop()

    def configure_from_args(self, args):
        vargs = vars(args)
        for arg in self.CONFIGURABLE_OPTIONS:
            if arg in vargs:
                setattr(self, arg, vargs[arg])

    def configure_from_env(self, prefix=""):
        for arg in self.CONFIGURABLE_OPTIONS:
            arg = prefix + arg.upper()
            if arg in os.environ:
                setattr(self, arg, os.environ[arg])

    def configure_from_docker_secrets(self):
        for arg in self.CONFIGURABLE_OPTIONS:
            spath = os.path.join('/run/secrets', arg)
            if os.path.exists(spath):
                value = open(spath).read().strip()
                setattr(self, arg, value)

    @classmethod
    def add_argparse_params(cls, parser):
        parser.add_argument("--mqtt-client-id", default="", required=False)
        parser.add_argument("--mqtt-username", default=None, required=False)
        parser.add_argument("--mqtt-password", default=None, required=False)
        parser.add_argument("--mqtt-host", default="localhost", required=False)
        parser.add_argument("--mqtt-port", default=1883, type=int, required=False)
        parser.add_argument("--mqtt-tls-cacert", default=None, required=False)
        parser.add_argument("--mqtt-tls-certfile", default=None, required=False)
        parser.add_argument("--mqtt-tls-keyfile", default=None, required=False)

        parser.add_argument("--discovery-prefix", default="homeassistant", required=False)
        parser.add_argument("--node-id", default=None, required=False)



import json
import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class Switch():
    """
    An MQTT switch
    """

    def __init__(self, name, entity_id):
        """
        :param name: The display name to use in HomeAssistant
        :param entity_id: The entity_id to use in HomeAssistant
        """
        self.name = name
        self.entity_id = entity_id

        self.discovery_prefix = None
        self.node_id = None
        self.client = None

        self._state = None

    @property
    def config(self):
        return {
            'name': self.name,
            'state_topic': self.state_topic,
            'command_topic': self.command_topic,
            'payload_on': self.payload_on,
            'payload_off': self.payload_off,
            'retain': self.retain
        }

    def connect(self, mqtt_client, discovery_prefix="homeassistant", node_id=None):
        """
        Connect this device to an MQTT broker

        :param mqtt_client: A connected MQTT client
        :param discovery_prefix: The discovery prefix set in the HomeAssistant config
        :param node_id: Will be included in the MQTT topics if set
        """
        self.discovery_prefix = discovery_prefix
        self.node_id = node_id
        self.client = mqtt_client

        self.client.message_callback_add(self.command_topic, self._on_command)
        self.client.subscribe(self.command_topic)

        self.client.publish(self.config_topic, json.dumps(self.config), retain=self.retain)
        logger.debug("Connected to broker, sent config to {}".format(self.config_topic))

    def _on_command(self, client, userdata, message):
        new_state = message.payload.decode('utf-8')
        logger.debug("Got command for new state {}, current state {}".format(new_state, self.state))
        if self._is_valid_state(new_state) and new_state != self.state:
            self.on_state_change(new_state)

    def _is_valid_state(self, state):
        return state in [self.payload_on, self.payload_off]

    def on_state_change(self, new_state):
        """
        Called when a state update request is recieved from the broker

        :param new_state: The state to set
        """
        self.state = new_state

    @property
    def state(self):
        """
        The state of the switch. Must be one of ``self.payload_on`` or ``self.payload_off``

        :getter: The last state the switch was set to (May be ``None`` if the state has never been set)
        :setter: Record the state change, and report it to the broker
        """
        return self._state

    @state.setter
    def state(self, value):
        if not self._is_valid_state(value):
            raise ValueError("New state must be one of {}, {}".format(self.payload_on, self.payload_off))

        self._state = value
        logger.debug("Publishing new state {}".format(value))
        self.client.publish(self.state_topic, value, retain=self.retain)

    @property
    def base_topic(self):
        if self.discovery_prefix is None:
            raise ValueError("Must call .connect() first")

        path = filter(lambda x: x is not None, [self.discovery_prefix, "switch", self.node_id, self.entity_id])
        return "/".join(path)

    @property
    def config_topic(self):
        return "/".join([self.base_topic, "config"])

    @property
    def state_topic(self):
        return "/".join([self.base_topic, "state"])

    @property
    def command_topic(self):
        return "/".join([self.base_topic, "command"])

    @property
    def payload_on(self):
        """
        Payload to use to indicate the switch is on. Defaults to ``"ON"``
        """
        return "ON"

    @property
    def payload_off(self):
        """
        Payload to use to indicate the switch is off. Defaults to ``"OFF"``
        """
        return "OFF"

    @property
    def retain(self):
        """
        Should the messages sent to the broker have the 'retain' flag set. Defaults to ``True``
        """
        return True



import pytest
from unittest import mock

import hassdevice
import hassdevice.devices
import hassdevice.hosts


class FakeMessage:
    def __init__(self, payload):
        self.payload = payload.encode('utf-8')


def test_switch_base_topic():
    with pytest.raises(ValueError):
        switch1 = hassdevice.devices.Switch("foo", "bar")
        switch1.base_topic

    switch2 = hassdevice.devices.Switch("foo", "bar")
    switch2.discovery_prefix = "homeassistant"
    assert switch2.base_topic == "homeassistant/switch/bar"

    switch3 = hassdevice.devices.Switch("foo", "bar")
    switch3.discovery_prefix = "homeassistant"
    switch3.node_id = "testserver"
    assert switch3.base_topic == "homeassistant/switch/testserver/bar"


@mock.patch('hassdevice.devices.Switch.on_state_change')
def test_state_change(mock_state_change):
    sw = hassdevice.devices.Switch("foo", "bar")
    sw._state = sw.payload_on
    sw._on_command(None, None, FakeMessage(sw.payload_off))
    assert mock_state_change.call_count == 1


@mock.patch('hassdevice.devices.Switch.on_state_change')
def test_state_change_no_update(mock_state_change):
    sw = hassdevice.devices.Switch("foo", "bar")
    sw._state = sw.payload_off
    sw._on_command(None, None, FakeMessage(sw.payload_off))
    assert mock_state_change.call_count == 0
